/**
 * Created by John on 7/5/2017.
 * Modified by TFG on 7/21/2020.
 *
 * Common logic for manipulating Loader models
 *
 * CAUTION !  each load now has a common starting base :
 *    csv_line[0] -> transactionId
 *    csv_line[1] -> file name
 *    csv_line[2] -> expected number of messages
 */
var express = require('express');
var app = express();

var awcblcommon = require('../config/awcblcommon');

var debug = true;               // Flag used to enable extra log messages

var moment = require("moment");
if (!String.prototype.startsWith) {
    String.prototype.startsWith = function(searchString, position) {
        position = position || 0;
        return this.indexOf(searchString, position) === position;
    };
}


function LoaderHelper(models, winston) {
    this.models = models;
    this.winston = winston
    this.logger = null;
    if (this.winston) {
        this.logger = this.winston.loggers.get("main");
    } else {
        this.logger = console;
        if (!this.logger.debug) this.logger.debug = console.log;
    }
}

LoaderHelper.prototype.loadCatalogs = function() {
    var self = this;
    // Loading vendor_list
    self.logger.debug("Loading Catalog table");

    self.models.CatalogList.findAll().then(function (catalogs) {
        if (catalogs) {
           if (debug === true) {
               self.logger.debug("loaded " + catalogs.length + " catalogs");
           }
           return catalogs;
        } else {
            if (debug === true) {
                self.logger.error("Error loading Catalog list");
            }
        }
    });
    return null;
}


/**
 * Saves a single Clinic Routing  record to the database tbl_vendor_route_clinic via promise
 * IMPORTANT ! It is assumed the data coming here are already vetted by m3-api during the
 * validation !!!!!!
 * It means no specific check on the data are performed here !
 * Expected CSV format :
 *   - transactionID
 *   - file_name
 *   - Number_messages
 *   - clinicId
 *   - vendorId
 *   - route
 *   - vendorDisplay
 *
 * @param {Object} msg JSON from Http
 */
LoaderHelper.prototype.saveLoaderCR = function(msg) {
    var self = this;
    var startingCSVindex = 3;

    return new Promise(function (resolve, reject) {
        var inClinicId;                 // Input clinic ID
        var inVendorId;                 // Input vendor ID
        var inPathRoute;                // Input Path route
        var inVendorDisplay;            // Input vendor display

        var csv_line = msg.split(/,(?=(?:(?:[^"]*"){2})*[^"]*$)/);
        csv_line = csv_line || [];

        if (csv_line.length > 1) {
            try {
                inClinicId = csv_line[startingCSVindex].split(',').join('');   // Assign input Clinic ID
                inVendorId = csv_line[startingCSVindex+1].split(',').join('');   // Assign input Vendor ID
                inPathRoute = csv_line[startingCSVindex+2].split(',').join('');   // Assign Path route
                // Extra parameter not in the original CSV file
                inVendorDisplay = csv_line[startingCSVindex+3].split(',').join('');   // Assign Vendor display name
            } catch (err) {
                // The test is needed in case the CSV record received is just the original one from the
                // bulk load CSV file, without the two extra parameters
                if (inVendorDisplay === undefined) {
                    inVendorDisplay = "--";
                } else {
                    reject("CAUTION !! CR - wrong CSV format received : " + err);
                }
            }

            if (debug === true) {
                self.logger.debug("  Payload retrieved ");
                self.logger.debug("           clinicID : " + inClinicId);  // Clinic ID
                self.logger.debug("           vendorId : " + inVendorId);  // Vendor Id
                self.logger.debug("          pathRoute : " + inPathRoute);  // Path route
                self.logger.debug("      vendorDisplay : " + inVendorDisplay);  // Vendor name
            }

            // TEMP ! Caution on the VendorDisplay

            self.models.VendorRouteClinic.findOrCreate({
                where: {
                    clinicId: inClinicId,
                    vendorId: inVendorId
                }, defaults: {
                    vendorDisplay: inVendorDisplay,
                    vendorPathRoute: inPathRoute
                    // no other parameters for default as the three above
                }
            }).spread(function (data, created) {
                if (!created) {
                    // Update record
//                    data.clinicId = inClinicId;
//                    data.vendorId = inVendorId;
                    data.vendorPathRoute = inPathRoute;

                    data.save().then(function () {
                        if (debug === true) {
                            self.logger.debug("Record updated ");
                        }
                        // Record updated successfully
                    }).catch(function (err) {
                        reject("CR - problem interacting with DB : " + err);
                    });
                } else {
                    // Record created successfully
                    if (debug === true) {
                        self.logger.debug("Record created ");
                    }
                }
                resolve("OK");
            }).catch(function (err) {
                reject("CR Error handling table : " + err);
            });
        } else {
            // Should never happens !!
            reject("CAUTION !! CR - something wrong received ! ");
        }
    });
};

/**
 * Saves a single Par Value record to the database tbl_sap_item_par_value via promise
 * IMPORTANT ! It is assumed the data coming here are already vetted by m3-api during the
 * validation !!!!!!
 * It means no specific check on the data are performed here !
 * Expected CSV format :
 *   - transactionID
 *   - file_name
 *   - Number_messages
 *   - clinicId
 *   - productId
 *   - qoh
 *   - min_par
 *   - max_par
 *   - vendorId
 *   - productName
 *
 * @param {Object} msg JSON from Http
 */
LoaderHelper.prototype.saveLoaderPV = function(msg) {
    var self = this;
    var startingCSVindex = 3;

    return new Promise(function (resolve, reject) {
        var inClinicId;                 // Input clinic ID
        var inProductId;                // Input product ID
        var inQoh;                      // Input QOH value
        var inMinParValue;              // Input Min par value
        var inMaxParValue;              // Input Max par value
        var inVendorId;                 // Input vendor ID
        var inProductName;              // Input product display name

        var csv_line = msg.split(/,(?=(?:(?:[^"]*"){2})*[^"]*$)/);
        csv_line = csv_line || [];

        if (csv_line.length > 1) {
            // Received something - check if is the information record
            try {
                // Prepare the fields to be loaded in the table reading the received record in the queue
                inClinicId = csv_line[startingCSVindex].split(',').join('');   // Assign input Clinic ID
                inProductId = csv_line[startingCSVindex+1].split(',').join('');   // Assign input Product ID
                inQoh = csv_line[startingCSVindex+2].split(',').join('');   // Assign input QOH
                inMinParValue = csv_line[startingCSVindex+3].split(',').join('');   // Assign input MinParValue
                inMaxParValue = csv_line[startingCSVindex+4].split(',').join('');   // Assign input MaxParValue
                // Extra parameters not in the original CSV file
                inVendorId = csv_line[startingCSVindex+5].split(',').join('');   // Assign input vendorId
                inProductName = csv_line[startingCSVindex+6].split(',').join('');   // Assign input Product Name
            } catch (err) {
                // The test is needed in case the CSV record received is just the original one from the
                // bulk load CSV file, without the two extra parameters
                if (inVendorId === undefined || inProductName === undefined) {
                    inVendorId = "--";
                    inProductName = "--";
                } else {
                    reject("CAUTION !! PV - wrong CSV format received : " + err);
                }
            }

            if (debug === true) {
                self.logger.debug("  Payload retrieved ");
                self.logger.debug("           clinicID : " + inClinicId);  // Clinic ID
                self.logger.debug("          productId : " + inProductId);  // Product ID
                self.logger.debug("                QOH : " + inQoh);  // QOH
                self.logger.debug("      min Par value : " + inMinParValue);  // Min
                self.logger.debug("      max Par value : " + inMaxParValue);  // Max
                self.logger.debug("           vendorId : " + inVendorId);  // Vendor Id
                self.logger.debug("        productName : " + inProductName);  // Product Name
            }

            self.models.VendorSapParValue.findOrCreate({
                where: {
                    vendorMaterial: inProductId,
                    vendorId: inVendorId,
                    clinicId: inClinicId
                }, defaults: {
                    qohValue: inQoh,
                    minParValue: inMinParValue,
                    maxParValue: inMaxParValue,
                    description: inProductName
                }
            }).spread(function (data, created) {
                if (!created) {
//                    data.vendorMaterial = inProductId;
//                    data.vendorId = inVendorId;
//                    data.clinicId = inClinicId;
                    data.qohValue = inQoh;
                    data.minParValue = inMinParValue;
                    data.maxParValue = inMaxParValue;
                    data.description = inProductName;

                    data.save().then(function () {
                        // Update successfull
                        if (debug === true) {
                            self.logger.debug("Record updated ");
                        }
                    }).catch(function (err) {
                        reject("PV - problem interacting with DB : " + err);
                    });
                } else {
                    // Creation successfull
                    if (debug === true) {
                        self.logger.debug("Record created ");
                    }
                }
                resolve("OK");
            }).catch(function (err) {
                reject("PV Error handling table : " + err);
            });
        } else {
            reject("CAUTION !! PV - something wrong received ! ");
        }
    });
};

/**
 * Saves a single Clinic Profile record to the database tbl_clinic_profile via promise
 * IMPORTANT ! It is assumed the data coming here are already vetted by m3-api during the
 * validation !!!!!!
 * It means no specific check on the data are performed here !
 * Expected CSV format :
 *   - transactionID
 *   - file_name
 *   - Number_messages
 *   - clinic_id
 *   - test_catalog
 *   - Mircera_visible
 *   - Mircera_Io
 *   - Areanesp_visible
 *   - Areanesp_Io
 *   - Epogen_Visible
 *   - Epogen_Io
 *   - Retacrit_visible
 *   - Retacrit_Io
 *   - Count_each_in
 *
 * @param {Object} msg JSON from Http
 */
LoaderHelper.prototype.saveLoaderCP = function(msg) {
    var self = this;
    var startingCSVindex = 3;

    return new Promise(function (resolve, reject) {
        var inClinicId;                 // Input clinic ID
        //var inTestCatalog;
        var inMirceraVisible;
        var inMirceraIo;
        var inAreanespVisible;
        var inAreanespIo;
        var inEpogenVisible;
        var inEpogenIo;
        var inRetacritVisible;
        var inRetacritIo;
        var inCountInEach;
        var inDateModified;

        var csv_line = msg.split(/,(?=(?:(?:[^"]*"){2})*[^"]*$)/);
        csv_line = csv_line || [];

        if (debug === true) {
            self.logger.debug("saveLoaderCP - csv line             : " + JSON.stringify(csv_line));
        }

        // NOTE !
        // The inTestCatalog is ALREADY a list of catalogs ID separated by the character #
        // Validation takes care also of the case of no catalog presents.
        // Has no use here since is used before to create the clinicCatalog list associated to this record

        if (csv_line.length > 1) {
            try {
                // Prepare the fields to be loaded in the table reading the received record in the queue
                inClinicId = csv_line[startingCSVindex].split(',').join('');
                //inTestCatalog = csv_line[startingCSVindex+1].split('"').join('');
                inMirceraVisible = csv_line[startingCSVindex+2].split('"').join('');
                inMirceraIo = csv_line[startingCSVindex+3].split('"').join('');
                inAreanespVisible = csv_line[startingCSVindex+4].split('"').join('');
                inAreanespIo = csv_line[startingCSVindex+5].split('"').join('');
                inEpogenVisible = csv_line[startingCSVindex+6].split('"').join('');
                inEpogenIo = csv_line[startingCSVindex+7].split('"').join('');
                inRetacritVisible = csv_line[startingCSVindex+8].split('"').join('');
                inRetacritIo = csv_line[startingCSVindex+9].split('"').join('');
                inCountInEach = csv_line[startingCSVindex+10].split('"').join('');
                inDateModified = new Date();
            } catch (e) {
                self.logger.error("Catch error : " + JSON.stringify(e));
                reject("CAUTION !! CP - wrong CSV format received ! ");
            }

            if (debug === true) {
                self.logger.debug("  Payload retrieved ");
                self.logger.debug("           clinicID : " + inClinicId);
                //self.logger.debug("      inTestCatalog : " + inTestCatalog);
                self.logger.debug("   inMriceraVisible : " + inMirceraVisible);
                self.logger.debug("        inMriceraIo : " + inMirceraIo);
                self.logger.debug("  inAreanespVisible : " + inAreanespVisible);
                self.logger.debug("       inAreanespIo : " + inAreanespIo);
                self.logger.debug("    inEpogenVisible : " + inEpogenVisible);
                self.logger.debug("         inEpogenIo : " + inEpogenIo);
                self.logger.debug("  inRetacritVisible : " + inRetacritVisible);
                self.logger.debug("       inRetacritIo : " + inRetacritIo);
                self.logger.debug("      inCountInEach : " + inCountInEach);
                self.logger.debug("     inDateModified : " + inDateModified);
            }

            // Building an object with the data received in input
            // Not saving on DB the vendorList (is a virtual field)
            var clinicProfile = self.models.ClinicProfile.build({
                clinicId: inClinicId,
                //vendorList: inTestCatalog,
                mirceraVisible: inMirceraVisible === 'y',
                mirceraIo: inMirceraIo === 'y',
                areanespVisible: inAreanespVisible === 'y',
                areanespIo: inAreanespIo === 'y',
                epogenVisible: inEpogenVisible === 'y',
                epogenIo: inEpogenIo === 'y',
                retacritVisible: inRetacritVisible === 'y',
                retacritIo: inRetacritIo === 'y',
                countInEach: inCountInEach === 'y',
                dateModified: inDateModified
            });

            self.models.ClinicProfile.findOrCreate({
                where: {
                    clinicId: inClinicId
                }, defaults: {
                    clinicProfile
                }
            }).spread(function (data, created) {
                if (!created) {
                    data.clinicId = clinicProfile.clinicId;
                    //data.vendorList = clinicProfile.vendorList;
                    data.mirceraVisible = clinicProfile.mirceraVisible;
                    data.mirceraIo = clinicProfile.mirceraIo;
                    data.areanespVisible = clinicProfile.areanespVisible;
                    data.areanespIo = clinicProfile.areanespIo;
                    data.epogenVisible = clinicProfile.epogenVisible;
                    data.epogenIo = clinicProfile.epogenIo;
                    data.retacritVisible = clinicProfile.retacritVisible;
                    data.retacritIo = clinicProfile.retacritIo;
                    data.countInEach = clinicProfile.countInEach;
                    data.dateModified = clinicProfile.dateModified;

                    data.save().then(function () {
                        self.logger.debug("   updated CP record " + data.id);

                        // TBD ? Update tbl_clinic_catalog
                    }).catch(function (err) {
                        reject("CP - problem interacting with DB : " + err);
                    });
                } else {
                    if (debug === true) {
                        self.logger.debug("   created CP record " + data.id);
                    }
                }
                resolve("OK");
            }).catch(function (err) {
                reject("CP Error handling table : " + err);
            });
        } else {
            reject("CAUTION !! CP - something wrong received ! ");
        }
    });
};

/**
 * Saves a an associated list of clinic catalog objects to the database tbl_clinic_catalog via promise
 * The input is an object of records
 *
 * @param {Object} msg
 */
LoaderHelper.prototype.saveLoaderCPassociated = function(clinicCatalogObject) {
    var self = this;

    var startingCSVindex = 3;

    return new Promise(function (resolve, reject) {
        if (debug === true) {
            self.logger.debug("saveLoaderCPassociated : " + JSON.stringify(clinicCatalogObject));
        }

        // Remove current associated records to the clinicId

        if (debug === true) {
            self.logger.debug("  --> remove clinic_id " + JSON.stringify(clinicCatalogObject[0].clinicId)  + " entries");
        }

        self.models.ClinicCatalog.destroy({
            where: {
                clinicId: clinicCatalogObject[0].clinicId
            }
        }).then(function (num) {
            // Removed entries
            // Save new list !

            var catpromises = [];

            cats.forEach(function (cat) {
                catpromises.push(cat.save());
            });

            self.models.sequelize.Promise.all(catpromises).then(function (vals) {
                return cats.length;
            }).catch(function (saveErr) {
                reject("CP associated Error - unable to save new data : " + saveErr);
            });
        }).catch(function (delErr) {
            reject("CP associated Error - unable to clean up old data : " + delErr);
        });


        // Save the associated list
//        for (var fn = 0; fn < clinicCatalogObject.length; fn++) {
//            var clinicCatalog = clinicCatalogObject[fn];

//            if (debug === true) {
//                self.logger.debug("  Payload to send ");
//                self.logger.debug("          clinicID : " + clinicCatalog.clinicId);
//                self.logger.debug("         catalogId : " + clinicCatalog.catalogId);
//            }

//            self.models.ClinicCatalog.findOrCreate({
//                where: {
//                    clinicId: clinicCatalog.clinicId
//                }, defaults: {
//                    catalogId: clinicCatalog.catalogId
//                }
//            }).spread(function (data, created) {
//                if (!created) {
//                    data.catalogId = clinicCatalog.catalogId;
//
//                    data.save().then(function () {
//                        self.logger.debug("   updated CP associated record " + data.id);
//                    }).catch(function (err) {
//                        reject("CP associated record - problem interacting with DB : " + err);
//                    });
//                } else {
//                    if (debug === true) {
//                        self.logger.debug("   created CP associated record " + data.id);
//                    }
//                }
//            }).catch(function (err) {
//                reject("CP Error handling table : " + err);
//            });
//        }

        resolve("OK");
    });
};

/**
 * Saves a single Master Item record to the database tbl_master_item via promise
 * IMPORTANT ! It is assumed the data coming here are already vetted by m3-api during the
 * validation !!!!!!
 * It means no specific check on the data are performed here !
 * Expected CSV format :
 *   - transactionID
 *   - file_name
 *   - Number_messages
 *
 * @param {Object} msg JSON from Http
 */
LoaderHelper.prototype.saveLoaderMT = function(msg) {
    var self = this;

    return new Promise(function (resolve, reject) {
        var inSapMaterial;
        var inDescription;
        var inCntInd;
        var inCntDescription;
        var inMedVendor;
        var inVendorMaterial;
        var inConv;
        var inCountType;
        var inAlternates;
        var inUnitPrice;
        var inVendorId;
        var inNdc;
        var inShowInventoryInd;
        var inDiscontinued;
        var inDateModified;

        var csv_line = msg.split(/,(?=(?:(?:[^"]*"){2})*[^"]*$)/);
        csv_line = csv_line || [];

        if (debug === true) {
            self.logger.debug("saveLoaderMT - csv line             : " + JSON.stringify(csv_line));
        }

        if (csv_line.length > 1) {
            try {
                // Prepare the fields to be loaded in the table reading the received record in the queue
                inSapMaterial = parseInt(line[0]);
                inDescription = csv_line[1].split('"').join('');
                inCntInd = csv_line[2].split('"').join('');
                inCntDescription = csv_line[3].split('"').join('');
                inMedVendor =  testCatalog;
                inVendorMaterial = csv_line[5].split('"').join('');
                inConv =  parseInt(csv_line[6].split('"').join(''));
                inCountType = csv_line[7].split('"').join('');
                inAlternates = csv_line[8].split('"').join('');
                inUnitPrice =  awcblcommon.formatPrice(app, csv_line[9].split('"').join(''));
                inVendorId = csv_line[10].split('"').join('');
                inNdc = csv_line[11].split('"').join('');
                inShowInventoryInd = csv_line[12].split('"').join('');
                inDiscontinued = awcblcommon.formatDate(app, csv_line[13].split('"').join(''));
                inDateModified =  new Date();
            } catch (e) {
                self.logger.error("Catch error : " + JSON.stringify(e));
                reject("CAUTION !! MT - wrong CSV format received ! ");
            }

            if (debug === true) {
                self.logger.debug("  Payload retrieved ");
                self.logger.debug("      inSapMaterial : " + inSapMaterial);
                self.logger.debug("      inDescription : " + inDescription);
                self.logger.debug("           inCntInd : " + inCntInd);
                self.logger.debug("   inCntDescription : " + inCntDescription);
                self.logger.debug("        inMedVendor : " + inMedVendor);
                self.logger.debug("   inVendorMaterial : " + inVendorMaterial);
                self.logger.debug("             inConv : " + inConv);
                self.logger.debug("        inCountType : " + inCountType);
                self.logger.debug("       inAlternates : " + inAlternates);
                self.logger.debug("        inUnitPrice : " + inUnitPrice);
                self.logger.debug("         inVendorId : " + inVendorId);
                self.logger.debug("              inNdc : " + inNdc);
                self.logger.debug(" inShowInventoryInd : " + inShowInventoryInd);
                self.logger.debug("     inDiscontinued : " + inDiscontinued);
                self.logger.debug("     inDateModified : " + inDateModified);
            }

            // Building an object with the data received in input

            var masterItem = self.models.MasterItem.build({
                sapMaterial: inSapMaterial,
                description: inDescription,
                cntInd: inCntInd === 'X',
                cntDescription: inCntDescription,
                medVendor: testCatalog,
                vendorMaterial: inVendorMaterial,
                conv: inConv,
                countType: inCountType,
                alternates: inAlternates,
                unitPrice: inUnitPrice,
                vendorId: inVendorId,
                ndc: inNdc,
                showInventoryInd: inShowInventoryInd === 'X',
                discontinued: inDiscontinued,
                dateModified: inDateModified
            });

            if (isNaN(masterItem.conv)) {
                masterItem.conv = 1;
            }

            self.models.MasterItem.findOrCreate({
                where: {
                    sapMaterial: masterItem.sapMaterial,
                    vendorMaterial: masterItem.vendorMaterial
                }, defaults: {
                    masterItem
                }
            }).spread(function (data, created) {
                if (!created) {
                    data.sapMaterial = masterItem.sapMaterial;
                    data.description = masterItem.description;
                    data.cntInd = masterItem.cntInd;
                    data.cntDescription= masterItem.cntDescription;
                    data.medVendor= testCatalog;
                    data.vendorMaterial= masterItem.vendorMaterial;
                    data.conv= masterItem.conv;
                    data.countType= masterItem.countType;
                    data.alternates= masterItem.alternates;
                    data.unitPrice= masterItem.unitPrice;
                    data.vendorId= masterItem.vendorId;
                    data.ndc= masterItem.ndc;
                    data.showInventoryInd= masterItem.showInventoryInd === 'X';
                    data.discontinued= masterItem.discontinued;
                    data.dateModified= masterItem.dateModified;

                    data.save().then(function () {
                        self.logger.debug("   updated MT record " + data.id);

                        // TBD ? Update tbl_clinic_catalog
                    }).catch(function (error) {
                        logger.error(error);
                    });
                } else {
                    self.logger.debug("   created MT record " + data.id);
                }
            }).catch(function (err) {
                reject(err);
            });
        } else {
            reject("CAUTION !! MT  - something wrong received ! ");
        }
    });
};

module.exports = LoaderHelper;
